"""Holz logging tool.

Utility functions to help with logging setup.
"""
# 2023-∞ (c) blurryroots innovation qanat OÜ. All rights reserved.
import logging


default_level = logging.WARNING
default_logger_name = 'holz'
default_logger = logging.getLogger (default_logger_name)
# All options at https://docs.python.org/3.8/library/logging.html#logrecord-attributes
default_format_str = u'[%(asctime)s][%(levelname)s][%(name)s]: %(message)s'
default_formatter = logging.Formatter (default_format_str)
default_overrides = {}
default_setup_silent = False
default_flush = False
logging.basicConfig (level = default_level, format = default_format_str)

ALWAYS_FLUSH = False
CURRENT_LOG_OVERRIDES = default_overrides


def configure (l, lvl, fmt, silent = False):
	'''Recreates output stream and configures given logger.'''
	# Avoid propagation and clear previous stream handlers.
	l.propagate = False
	l.handlers.clear ()

	# Change logger level.
	if not silent:
		default_logger.debug (f'Setting {l} to level {lvl}')
	l.setLevel (lvl)

	# Create streaming log channler and set level.
	ch = logging.StreamHandler ()
	ch.setLevel (lvl)
	ch.setFormatter (fmt)
	ch.set_name(f'{l.name}:StreamHandler')
	l.addHandler (ch)

	return l


def _handle_logger_config_override (overrides, l, silent = default_setup_silent):
	'''Utility function to override logger specific configurations.'''
	lvl = overrides[l.name]['level']
	fmttr = default_formatter
	if 'formatter' in overrides[l.name] \
		and not (overrides[l.name]['formatter'] is None \
		):
		fmttr = overrides[l]['formatter']
	configure (l, lvl, fmttr, silent = silent)


def setup (log_name
	, log_level = default_level
	, log_formatter = default_formatter
	, silent = False
):
	'''Create and configure a specific logger.'''
	logger = logging.getLogger (log_name)
	logger = configure (logger
		, log_level
		, log_formatter
		, silent = silent
	)

	return logger


def activate_flush_always (flush):
	'''Indicate holz should flush logging output always.'''
	global ALWAYS_FLUSH
	ALWAYS_FLUSH = flush


def setup_default (log_name
	, log_level = default_level
	, log_formatter = default_formatter
	, log_overrides = {}
	, silent = False
):
	'''Configure holz default values.'''
	global default_logger_name
	global default_logger
	global default_level
	global default_formatter
	global default_overrides
	global default_setup_silent

	if not silent:
		default_logger.debug (f'Changing default log to "{log_name}".')

	default_logger_name = log_name
	default_level = log_level
	default_formatter = log_formatter
	default_overrides = log_overrides
	default_setup_silent = silent
	default_logger = setup (default_logger_name
		, default_level
		, default_formatter
		, default_setup_silent
	)

	return default_logger


def normalize (overrides = {}, silent = default_setup_silent):
	'''Looks for all known logger and normalizes their output and level.'''
	global CURRENT_LOG_OVERRIDES

	if CURRENT_LOG_OVERRIDES != overrides:
		CURRENT_LOG_OVERRIDES = overrides

	# collect all logges and normalize their setup / configuration
	loggers = [logging.getLogger ()] \
		+ [logging.getLogger (name) for name in logging.root.manager.loggerDict]

	for l in loggers:
		if not silent:
			default_logger.debug (f'Normalizing setup for logger {l} ...')
		if l.name in CURRENT_LOG_OVERRIDES:
			if not silent:
				default_logger.debug (f'Found override for "{l}". Applying ...')
			_handle_logger_config_override (CURRENT_LOG_OVERRIDES
				, l
				, silent = silent
			)
		else:
			configure (l, default_level, default_formatter, silent = silent)


def _log (method, message, category, flush = default_flush):
	'''Fetches logger and outputs message with level-specific output.'''
	logger = None
	if category:
		logger = logging.getLogger (category)
	else:
		logger = default_logger
	m = getattr (logger, method)
	m (message)
	if flush or ALWAYS_FLUSH:
		for h in logger.handlers:
			h.flush ()


def debug (message, category = None, flush = False):
	_log ('debug', message, category, flush)


def info (message, category = None, flush = False):
	_log ('info', message, category, flush)


def warn (message, category = None, flush = False):
	_log ('warning', message, category, flush)


def error (message, category = None, flush = False):
	_log ('error', message, category, flush)


def fatal (message, category = None, flush = False):
	_log ('fatal', message, category, flush)


class HolzNotifyLogger(logging.Logger):
	"""
	Logger subclass to notify holz when a new logger is created / requested.
	"""
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		global ON_LOGGER_REQUESTED
		
		if self.name in CURRENT_LOG_OVERRIDES:
			_handle_logger_config_override(CURRENT_LOG_OVERRIDES
				, self
				, default_setup_silent
			)


logging.setLoggerClass(HolzNotifyLogger)



"""CLI entry point.

CLI handling of piper-whistle commands.
"""
# 2023-∞ (c) blurryroots innovation qanat OÜ. All rights reserved.
import os
import sys
import pathlib
import argparse
import logging
# Append root package to path so it can be called with absolute path.
sys.path.append (str (pathlib.Path(__file__).resolve().parents[1]))
# ..
from piper_whistle import holz
from piper_whistle import db
from piper_whistle import cmds
from piper_whistle import util
from piper_whistle import version


# Build command lookup map.
commands = {
	'refresh': cmds.run_refresh,
	'guess': cmds.run_guess,
	'path': cmds.run_path,
	'speak': cmds.run_speak,
	'list': cmds.run_list,
	'preview': cmds.run_preview,
	'install': cmds.run_install,
	'remove': cmds.run_remove
}
HELP_REQUESTED = False


class WhistleArgsParserException (Exception):
	'''.'''
	LOG = logging.getLogger ('whistle-args-exception')
	
	def __init__ (self, status, message, help_requested = False):
		self.LOG.debug (
			f'Creating new (status: {status}), '
			f'message: "{message}", '
			f'help_requested: {help_requested}'
		)
		super ().__init__ (message)
		self.status = status
		self.help_requested = help_requested

	@property
	def message (self):
		return str (self)


class WhistleArgsParser (argparse.ArgumentParser):
	'''.'''
	LOG = logging.getLogger ('whistle-args')

	def __init__ (self, *args, **kwargs):
		'''.'''
		n = kwargs["prog"] if "prog" in kwargs else "UNNAMED"
		self.LOG.debug (
			f'Setting up "{n}" arg parser.'
		)
		super ().__init__ (*args, **kwargs)
		self._last_error_code = 0
		self._last_error_message = None
		self._help_requested = False

	def print_help (self):
		'''.'''
		self._help_requested = True
		self.LOG.debug ('Printing help, then raising WhistleArgsParserException.')
		# Idea based on discussion at:
		# https://stackoverflow.com/a/61039719
		# First print auto-generated help text, then instead of exiting,
		# raise whistle exception so main function can react. It smells like
		# 'exception as flow control' anti-pattern'ish, yet it apperas to me
		# as a reasonable trait-off for this scenario.
		super ().print_help ()
		self.LOG.debug ('Here we go ...')
		raise WhistleArgsParserException (
			self._last_error_code,
			self._last_error_message,
			help_requested=True
		)

	def print_help_raw (self):
		'''.'''
		super ().print_help ()

	def error (self, message):
		'''.'''
		self._last_error_message = message
		self._last_error_code = 13
		raise WhistleArgsParserException (
			self._last_error_code,
			self._last_error_message
		)

	def exit (self, status=0, message=None):
		'''.'''
		self._last_error_code = status
		self._last_error_message = message

	def has_error (self):
		'''.'''
		return 0 < self._last_error_code

	def last_error (self):
		'''.'''
		return self._last_error_code, self._last_error_message


def create_arg_parser (prog = 'piper_whistle'):
	"""! Build argparse command line argument parser."""

	# Build top level parser object.
	parser = WhistleArgsParser (prog = prog
		, formatter_class=argparse.RawTextHelpFormatter
		, add_help=False
	)

	# Setup global flags for verbosity level and version print.
	parser.add_argument ('-h', '--help'
		, action='help'
		, help='Show help message.'
		, default=False
	)
	parser.add_argument ('-d', '--debug'
		, action='store_true'
		, help='Activate very verbose logging.'
		, default=False
	)
	parser.add_argument ('-v', '--verbose'
		, action='store_true'
		, help='Activate verbose logging.'
		, default=False
	)
	parser.add_argument ('-V', '--version'
		, action='store_true'
		, help='Show version number.'
		, default=False
	)
	parser.add_argument ('-P', '--data-root', type=str
		, help='Root path where whistle should store config and data in.'
		, default=None
	)
	parser.add_argument ('-R', '--refresh'
		, action='store_true'
		, help='Refreshes (or sets up) language index by downloading the latest lookup.'
		, default=False
	)

	# Split object for subparsers.
	subparsers = parser.add_subparsers (dest='command')

	# Setup refresh command and options.
	refresh_args = subparsers.add_parser ('refresh'
		, formatter_class=argparse.RawTextHelpFormatter
		, add_help=False
	)
	refresh_args.add_argument ('-h', '--help'
		, action='help'
		, help='Show help message.'
		, default=False
	)
	refresh_args.add_argument ('-v', '--verbose'
		, action='store_true'
		, help='Activate verbose logging.'
		, default=False
	)
	refresh_args.add_argument ('-R', '--repository'
		, type=str
		, help=\
			'Configures the huggingface repository. Will be used to build index,'
			' and download all data (e.g. models) from.'
		, default='rhasspy/piper-voices'
	)

	# Setup gues command and options.
	guess_args = subparsers.add_parser ('guess'
		, formatter_class=argparse.RawTextHelpFormatter
		, add_help=False
	)
	guess_args.add_argument ('-h', '--help'
		, action='help'
		, help='Show help message.'
		, default=False
	)
	guess_args.add_argument ('-v', '--verbose'
		, action='store_true'
		, help='Activate verbose logging.'
		, default=False
	)
	guess_args.add_argument ('language_name', type=str
		, help='A string representing a language name (or code).'
		, default=''
	)

	# Setup path command and options.
	selector_args = subparsers.add_parser ('path'
		, formatter_class=argparse.RawTextHelpFormatter
	)
	selector_args.add_argument ('-v', '--verbose'
		, action='store_true'
		, help='Activate verbose logging.'
		, default=False
	)
	selector_args.add_argument ('voice_selector', type=str
		, help='Selector of voice to search.'
		, default=''
	)

	# Setup speak command and options.
	speak_args = subparsers.add_parser ('speak'
		, formatter_class=argparse.RawTextHelpFormatter
		, add_help=False
	)
	speak_args.add_argument ('-h', '--help'
		, action='help'
		, help='Show help message.'
		, default=False
	)
	speak_args.add_argument ('something', type=str
		, help='Something to speak.'
		, default=''
	)
	speak_args.add_argument ('-c', '--channel'
		, type=str
		, help='Path to channel (named pipe (aka. fifo)) to which piper is listening.'
		, default='/opt/wind/channels/speak'
	)
	speak_args.add_argument ('-j', '--json'
		, action='store_true'
		, help='Encode the text as json payload. Is on by default.'
		, default=True
	)
	speak_args.add_argument ('-r', '--raw'
		, action='store_true'
		, help='Encode the text directly.'
		, default=False
	)	
	speak_args.add_argument ('-o', '--output'
		, type=str
		, help=\
			'Instead of streaming to audio channel, specifies a path to wav'
			' file where speech will be store in.'
		, default=None
	)
	speak_args.add_argument ('-v', '--verbose'
		, action='store_true'
		, help='Activate verbose logging.'
		, default=False
	)

	# Setup list command and options.
	list_args = subparsers.add_parser ('list'
		, formatter_class=argparse.RawTextHelpFormatter
		, add_help=False
	)
	list_args.add_argument ('-h', '--help'
		, action='help'
		, help='Show help message.'
		, default=False
	)
	list_args.add_argument ('-v', '--verbose'
		, action='store_true'
		, help='Activate verbose logging.'
		, default=False
	)
	list_args.add_argument ('-I', '--installed'
		, action='store_true'
		, help='Only list installed voices.'
		, default=False
	)
	list_args.add_argument ('-a', '--all'
		, action='store_true'
		, help='List voices for all available languages.'
		, default=False
	)
	list_args.add_argument ('-L', '--languages'
		, action='store_true'
		, help='List available languages.'
		, default=False
	)
	list_args.add_argument ('-g', '--legal'
		, action='store_true'
		, help='Show avaiable legal information.'
		, default=False
	)
	list_args.add_argument ('-U', '--show-url'
		, action='store_true'
		, help='Show URL of voice on remote host.'
		, default=False
	)
	list_args.add_argument ('-S', '--omit-speakers'
		, action='store_true'
		, help='Omit speakers form listing.'
		, default=False
	)
	list_args.add_argument ('-p', '--install-path'
		, action='store_true'
		, help='Show path of voice (if installed).'
		, default=False
	)
	list_args.add_argument ('-l', '--language-code'
		, type=str
		, help='Only list voices matching this language.'
		, default='en_GB'
	)
	list_args.add_argument ('-i', '--voice-index'
		, type=int
		, help='List only specific language voice.'
		, default=-1
	)

	# Setup preview command and options.
	preview_args = subparsers.add_parser ('preview'
		, formatter_class=argparse.RawTextHelpFormatter
		, add_help=False
	)
	preview_args.add_argument ('-h', '--help'
		, action='help'
		, help='Show help message.'
		, default=False
	)
	preview_args.add_argument ('-v', '--verbose'
		, action='store_true'
		, help='Activate verbose logging.'
		, default=False
	)
	preview_args.add_argument ('-l', '--language-code'
		, type=str
		, help='Select language.'
		, default='en_GB'
	)
	preview_args.add_argument ('-i', '--voice-index'
		, type=int
		, help='Specific language voice. (defaults to first one)'
		, default=0
	)
	preview_args.add_argument ('-s', '--speaker-index'
		, type=int
		, help='Specific language voice speaker. (defaults to first one)'
		, default=0
	)
	preview_args.add_argument ('-D', '--dry-run'
		, action='store_true'
		, help='Build URL and simulate download.'
		, default=False
	)
	
	# Setup install command and options.
	install_args = subparsers.add_parser ('install'
		, formatter_class=argparse.RawTextHelpFormatter
		, add_help=False
	)
	install_args.add_argument ('-h', '--help'
		, action='help'
		, help='Show help message.'
		, default=False
	)
	install_args.add_argument ('-v', '--verbose'
		, action='store_true'
		, help='Activate verbose logging.'
		, default=False
	)
	install_args.add_argument ('-D', '--dry-run'
		, action='store_true'
		, help='Simulate download / install.'
		, default=False
	)
	install_args.add_argument ('language_code'
		, type=str
		, help='Select language.'
	)
	install_args.add_argument ('voice_index'
		, type=int
		, help='Specific language voice. (defaults to first one)'
	)

	# Setup remove command and options.
	remove_args = subparsers.add_parser ('remove'
		, formatter_class=argparse.RawTextHelpFormatter
		, add_help=False
	)
	remove_args.add_argument ('-h', '--help'
		, action='help'
		, help='Show help message.'
		, default=False
	)
	remove_args.add_argument ('-v', '--verbose'
		, action='store_true'
		, help='Activate verbose logging.'
		, default=False
	)
	remove_args.add_argument ('voice_selector', type=str
		, help='Selector of voice to search.'
		, default=''
	)

	return parser


def main (custom_args, force_debug = False):
	"""! Main CLI processing function."""
	log_level = logging.WARNING

	# Setup and configure argparse parser.
	parser_ex = None
	parser = create_arg_parser ()
	try:
		# Parse passed arguments.
		args = parser.parse_args (args = custom_args[1:])
	except Exception as ex:
		# Setup holz just before handling exception.
		if force_debug:
			log_level = logging.DEBUG
		holz.setup_default ('whistle', log_level)
		holz.normalize ()
		# Remember exception.
		parser_ex = ex

	if None is parser_ex:
		holz.debug ('Parsed arguments without exception.')
	elif type (parser_ex) == WhistleArgsParserException:
		# So here we are, exception as flow control. It's just very convenient
		# to be able to identify help requested, since otherwise it seems
		# I'd have to equip every sub-arg-parser with a reference to the main
		# parser and then notify when help was requested in any sub-command.
		# So it comes down to lazyness?! Or economy? Anyway, it seems reasonable.
		if parser_ex.help_requested:
			holz.debug ('Parser help was requested.')
			# Since message has already been put to stdout, we can exit.
			return 0

		# Notify about exception.
		holz.error (parser_ex.message)

		# Return forwarded status code.
		return parser_ex.status
	else:
		holz.fatal (f'Unexpected error: {parser_ex}')
		return 23


	# Determine log level for holz setup.
	if args.debug or force_debug:
		log_level = logging.DEBUG
	elif args.verbose:
		log_level = logging.INFO

	# Setup default log level and initialize holz logging utility.
	holz.setup_default ('whistle', log_level)
	holz.activate_flush_always (True)
	holz.normalize ()
	holz.debug ('Holz setup done and available loggers normalized.')

	if args.version:
		print (version.as_string ())
		return 0

	if args.help:
		# 'Print only function', since overridden print_help raises an exception.
		parser.print_help_raw ()
		return 0

	# Fetch default paths for config and data storage.
	paths = None
	if args.data_root:
		if not os.path.exists (args.data_root):
			holz.error (f'Data path not found! ({args.data_root})')
			return 1
		paths = db.data_paths (args.data_root)
	else:
		paths = db.data_paths ()

	# Fetch details on where to obtain voice data from.
	repo_info = db.remote_repo_config (paths)

	# Trying to create new context object. Might fail if database is missing / corrupt.
	context = db.context_create (paths, repo_info)
	if not context:
		# Check if refresh is requestsed.
		if args.refresh or 'refresh' == args.command:
			holz.debug ('Refresh requested. Manually constructing context ...')
			context = {
				'paths': paths,
				'db': None,
				'repo': repo_info
			}
		# Otherwise cya.
		else:
			holz.error (
				f'Could not create context. ' \
				f'Please refresh database using "{sys.argv[0]} refresh"'
			)
			# 'Print only function', since overridden print_help raises an exception.
			parser.print_help_raw ()
			return 1

	# Show help message if no command is provided.
	if None is args.command:
		holz.debug ('No command specified.')
		# 'Print only function', since overridden print_help raises an exception.
		parser.print_help_raw ()
		return 1

	# Select command and run it.
	if args.command in commands:
		holz.debug (f'Running command "{args.command}" ...')
		r = commands[args.command] (context, args)
		holz.debug (f'Finished with "{r}".')
		return r

	holz.error (f'Could not find action binding for command "{args.command}".')

	# Show available commands.
	# Use 'print only function', since overridden print_help raises an exception.
	parser.print_help_raw ()
	return 1


def main_sys ():
	"""! CLI entry point with sys.argv as default parameters."""
	# Call CLI entry function.
	r = main (sys.argv)
	# Exit with return code.
	sys.exit (r)


if __name__ == '__main__':
	"""! CLI entry point."""
	main_sys ()



"""Search tools

Utilities for fuzzy searching / comparison.
"""
# 2023-∞ (c) blurryroots innovation qanat OÜ. All rights reserved.
from thefuzz import fuzz as fuzz_core


def looks_like (needle, haystack, threshold = 0.66, case_sensitive = True):
	"""! Fuzzy comparison / containment check.

	Takes a needle and searches haystack for the closest partial match given
	a specific threshold and case-mode.

	@param needle Search phrase.
	@param haystack	Corpuse to search in.
	@param threshold Match coefficient [0,1] Higher values force stricter closeness.
	@param case_sensitive Wheter to care for case differences.

	@return Returns a touple of whether match occurs, and the confidence coefficient.
	"""
	if not case_sensitive:
		needle = needle.lower ()
		haystack = haystack.lower ()

	confidence = fuzz_core.partial_ratio (needle, haystack)
	confidence /= 100.0

	return threshold < confidence, confidence



"""Whistle database functionallity.

Defines functions to interact with the meta information / database of whistle.
All relevant files and paths can be accessed via "data_paths ()"

The index is refreshed by querying the hugging face repository "rhasspy/piper-voices"

Here's an example of how to identify different parts of a voice via URL

URL: https://huggingface.co/rhasspy/piper-voices/resolve/main/pt/pt_PT/tug%C3%A3o/medium/pt_PT-tug%C3%A3o-medium.onnx.json

	* root: https://huggingface.co/rhasspy/piper-voices
	* op: /resolve
	* branch: /main
	* code_short: /pt
	* code_long: /pt_PT
	* voice_name: /tug%C3%A3o
	* voice_quality: /medium
	* voice_config: /pt_PT-tug%C3%A3o-medium.onnx.json

"""
# 2023-∞ (c) blurryroots innovation qanat OÜ. All rights reserved.
import os
import sys
import json
import userpaths
import time
import urllib
import pathlib
import tempfile
# Append root package to path so it can be called with absolute path.
sys.path.append (str (pathlib.Path(__file__).resolve().parents[1]))
# ..
from piper_whistle import holz
from piper_whistle import util
from piper_whistle import search


def data_paths (appdata_root_path = userpaths.get_appdata ()):
	"""! Query for data paths used by whistle.

	Provides a map of data paths used by whistle.
	
	* data: Root path of whistle data.
	* voices: Storage path for voice data.
	* index: Language details cached from huggingface repository (JSON).
	* languages: Language data lookup (JSON), built from index. Uses language code as keys.
	* last-updated: A flat file containig the timestamp when whistle data was refreshed last.
	
	@param config_root_path Path to the directory, where applications can store configuration and data.
	@return Returns a map with respective paths.
	"""
	whistle_data_path = pathlib.Path (appdata_root_path).joinpath ('piper-whistle')
	return {
		'data': whistle_data_path.as_posix (),
		'repo': whistle_data_path.joinpath ('repo.json').as_posix (),
		'voices': whistle_data_path.joinpath ('voices').as_posix (),
		'index': whistle_data_path.joinpath ('index.json').as_posix (),
		'languages': whistle_data_path.joinpath ('languages.json').as_posix (),
		'last-updated': whistle_data_path.joinpath ('last-updated').as_posix ()
	}


def remote_repo_config (paths):
	"""! Returns information about the remote repository.

	The map return comes with the following fields.

	root: Repository root url (e.g. huggingface).
	repo-id: Repository identifier (in format "username/reponame")
	branch: Branch name,
	voice-index: Path to voice index file (JSON). Relative to repository root.

	@return Returns a map with relevant repository information.
	"""
	repo = None
	repo_config_file = pathlib.Path (paths['repo'])
	if not repo_config_file.exists ():
		repo = {
			'root': 'https://huggingface.co',
			'repo-id': 'rhasspy/piper-voices',
			'branch': 'main',
			'voice-index': 'voices.json'
		}
		holz.debug (f'Writing default repo config to "{repo_config_file}" ...')
		# Make sure data root path exists, in case repo config is the first to use it.
		repo_config_file.parent.mkdir (parents = True, exist_ok = True)
		with open (repo_config_file.as_posix (), 'w') as f:
			json.dump (repo, f, indent = 4)

	holz.debug (f'Loading repo config from "{repo_config_file}" ...')
	with open (repo_config_file.as_posix (), 'r') as f:
		repo = json.load (f)

	return repo


def remote_repo_build_branch_root (repo_info):
	"""! Build repository root URL.
	
	The base URL will point to the root of the repository, including the branch.

	@param repo_info Remote repo information map. Can be obtained via @ref "remote_repo_config ()".
	@return Returns a list of file info maps with keys {path,size} or None on error.
	"""
	repo_root = f"{repo_info['root']}/{repo_info['repo-id']}"
	branch_root = f"{repo_root}/resolve/{repo_info['branch']}"

	return branch_root


def remote_repo_build_index_url (repo_info):
	"""! Build URL pointing to repository voice index file.

	This will produce the link to the voices.json index file, which will be used
	to build whistle database and lookup.

	@param repo_info Remote repo information map. Can be obtained via @ref "remote_repo_config ()".
	@return Returns a list of file info maps with keys {path,size} or None on error.
	"""
	branch_root = remote_repo_build_branch_root (repo_info)
	index_url = f"{branch_root}/{repo_info['voice-index']}"

	return index_url


def assemble_download_info (context, code, voice_i):
	"""! Compile details used to download voice data.

	The assembled information may look like this:
	{
		'langugage': en_GB,
		'model': {
			'url': "https://...",
			'size': 777,
			'md5': some md5 hash
		},
		'config': {
			'url': "https://...",
			'size': 777,
			'md5': some md5 hash
		},
		'samples': [
			list of URLs to a sample voice reading for each speaker
		],
		'local_path_relative': /some/local/path,
		'selection_name': selector name,
	}	

	@param context Context information and whistle database.
	@param code Language code of voice to be downloaded.
	@param voice_i Voice index to be downloaded.
	@return Returns a map containing download information.
	"""
	index = context['db']['index']
	langdb = context['db']['languages']
	# For example: "https://huggingface.co/rhasspy/piper-voices/resolve/main"
	base_url = remote_repo_build_branch_root (context['repo'])

	# Check if given code is available and collect meta info for later
	# download and storage.
	if code in langdb:
		lang = langdb[code]
		voices = lang["voices"]
		# Select specific voice by index.
		if -1 < voice_i and voice_i < len (voices):
			voice_name = voices[voice_i]
			holz.info (f'Requesting "{voice_name}" ...')
			voice_details = index[voice_name]
			
			download_info = {
				'langugage': code,
				'model': None,
				'config': None,
				'card': None,
				'samples': [],
				'local_path_relative': f'{code}/{voice_name}',
				'selection_name': f"{code}:{voice_details['name']}@{voice_details['quality']}"
			}

			# Identify onnx speech model files.
			for file in voice_details['files']:
				if file.endswith ('.onnx'):
					download_info['model'] = {
						'url': f'{base_url}/{file}',
						'size': voice_details['files'][file]['size_bytes'],
						'md5': voice_details['files'][file]['md5_digest']
					}
				elif file.endswith ('.onnx.json'):
					download_info['config'] = {
						'url': f'{base_url}/{file}',
						'size': voice_details['files'][file]['size_bytes'],
						'md5': voice_details['files'][file]['md5_digest']
					}
				elif file.endswith ('MODEL_CARD'):
					download_info['card'] = {
						'url': f'{base_url}/{file}',
						'size': voice_details['files'][file]['size_bytes'],
						'md5': voice_details['files'][file]['md5_digest']
					}

			voice_base_url = os.path.dirname (download_info['model']['url'])

			def build_sample_url (base, speaker_name, speaker_id, ext = 'mp3'):
				return f'{base}/samples/{speaker_name}_{speaker_id}.{ext}'

			# samples are based on speakers.
			# there is always a speaker 0 by default.
			if 1 >= int (voice_details['num_speakers']):
				speaker_url = build_sample_url (voice_base_url, 'speaker', 0)
				download_info['samples'].append (speaker_url)
			else:
				for speaker_name in voice_details['speaker_id_map']:
					speaker_id = int(voice_details['speaker_id_map'][speaker_name])
					speaker_url = build_sample_url (voice_base_url, 'speaker', speaker_id)
					download_info['samples'].append (speaker_url)

			return download_info
		else:
			holz.error (f'Invalid voice index!')
	else:
		holz.error (f'Cannot recognize: "{code}"')

	return None


def _fetch_url_raw (url):
	"""
	"""
	temp = tempfile.NamedTemporaryFile ()
	temp_path = temp.name
	temp.close ()

	r = util.download_as_stream_with_progress (url, temp_path)
	if 0 < r:
		holz.debug (f'Finished downloading. "{url}" => "{temp_path}"')
		with open (temp_path, 'r') as f:
			return f.read ()

	return None


def index_fetch_raw_filelist (repo_info):
	"""! Query the huggingface repository for a list of model files.
	@param repo_info Remote repo information map. Can be obtained via @ref "remote_repo_config ()".
	@return Returns a list of file info maps with keys {path,size} or None on error.
	"""
	import huggingface_hub as hf

	api = hf.HfApi ()

	repo_id = repo_info['repo-id']
	r = api.model_info (repo_id, files_metadata = True)
	if 300 > r.status_code:
		filelist = []
		for file in r.siblings:
			filelist.append ({ 'path': file.rfilename, 'size': file.size })

		return filelist
	
	holz.error (f'Could not fetch raw list from "{repo_id}".')
	return None


def index_fetch_raw (repo_info):
	"""! Query the huggingface repository for most recent voice index.
	@param repo_info Remote repo information map. Can be obtained via @ref "remote_repo_config ()".
	@return Returns parse json object of voice index.
	"""
	url = remote_repo_build_index_url (repo_info)

	raw_text = _fetch_url_raw (url)
	if raw_text:
		return json.loads (raw_text)

	holz.error (f'Could note fetch index.')
	return None


def _parse_model_card (card_text):
	lines = [line.strip () for line in card_text.split ('\n')]
	entry = {
		'dataset-url': None,
		'license': None,
		'training': None,
		'reference': None
	}

	read_training_status = False
	ln = 0
	for line in lines:
		if 0 == len (line):
			continue

		if read_training_status:
			l = line.lower ()
			if l.startswith ('finetuned'):
				entry['training'] = 'Tuned'
				entry['reference'] = line.replace ('Finetuned from', '').strip ()
			elif l.startswith ('fine-tuned'):
				entry['training'] = 'Tuned'
				entry['reference'] = line.replace ('Fine-tuned from', '').strip ()
			else:
				entry['training'] = 'Original'
				entry['reference'] = line.replace ('Trained from scratch', '')
				if entry['reference'].endswith ('.'):
					entry['reference'] = entry['reference'][:-1]
			read_training_status = False
		elif line.startswith ('* License:'):
			entry['license'] = line.replace ('* License:', '').strip ()
		elif line.startswith ('* URL:'):
			entry['dataset-url'] = line.replace ('* URL:', '').strip ()
		elif line.startswith ('## Training'):
			read_training_status = True
		ln += 1

	return entry


def index_download_and_rebuild (paths, repo_info):
	"""! Fetch latest voice index and build lookup database, then recreates context.

	Using the info on data paths and remote repository, it fetches the latest
	voice information, re-builds the cache / database, creates a context
	and returns it.
	
	@param context Context map containig whistle index and language info.
	               Can be created via @ref "context_create ()".
	
	@return Returns parse json object of voice index.
	"""
	holz.debug ('Ensuring data paths exists ...')

	# Make sure data path exists.
	pathlib.Path (paths['data']).mkdir (parents = True, exist_ok = True)

	# Make sure voice data storage path exists.
	pathlib.Path (paths['voices']).mkdir (parents = True, exist_ok = True)

	holz.info ('Fetching current index ...')
	index = index_fetch_raw (repo_info)
	if not index:
		holz.error ('Unable to build index.')
		return None

	holz.info (f"Storing index at '{paths['index']}' ...")
	with open (paths['index'], 'w') as f:
		json.dump (index, f, indent = 4)

	holz.info ('Rebuilding language database ...')
	langdb = {}
	for voice in index:
		l = index[voice]['language']
		if not (l['code'] in langdb):
			langdb[l['code']] = l
			langdb[l['code']]['voices'] = []
		langdb[l['code']]['voices'].append (index[voice]['key'])
	
	with open (paths['languages'], 'w') as f:
		json.dump (langdb, f, indent = 4)

	holz.info ('Updating timestamp ...')
	last_update = time.time ()
	with open (paths['last-updated'], 'w') as f:
		f.write (f'{last_update}')

	holz.info (f"Database files stored at '{paths['data']}'.")

	# build legal info based on model cards
	legal = {}
	if True:
		base_url = remote_repo_build_branch_root (repo_info)

		for code in langdb:
			holz.info (f'Processing "{code}":')				
			voice_i = 0
			for voice_name in langdb[code]['voices']:
				holz.info (f"\tFetching model card for {voice_i}: {voice_name}")
				dl_info = None
				voice_details = index[voice_name]
				# Identify onnx speech model files.
				for file in voice_details['files']:
					if file.endswith ('MODEL_CARD'):
						dl_info = {
							'url': f'{base_url}/{file}',
							'size': voice_details['files'][file]['size_bytes'],
							'md5': voice_details['files'][file]['md5_digest']
						}

				model_card_text = _fetch_url_raw (dl_info['url'])
				card = _parse_model_card (model_card_text)

				holz.info (f'\tStoring license ...')
				legal[voice_name] = card

				voice_i = voice_i + 1

		with open (os.path.join (paths['data'], 'legal.json'), 'w') as f:
			json.dump (legal, f, indent = 4)

	holz.info ('Regenerating context ...')
	context = context_create (paths, repo_info)

	"""TODO: consider automatic lookup / best guess of corpus data
	# corpus licenses lookup
	# https://raw.githubusercontent.com/coqui-ai/open-speech-corpora/master/README.md
	corpus_lookup_raw = _fetch_url_raw ('https://raw.githubusercontent.com/coqui-ai/open-speech-corpora/master/README.md')
	if corpus_lookup_raw:
		with open (os.path.join (paths['data'], 'corpus-lookup.md'), 'w') as f:
			f.write (corpus_lookup_raw)
	"""
	return context


def _context_is_valid (context):
	paths_ok = ('paths' in context)
	if not paths_ok:
		holz.warn ('Data paths appear corrupt.')
	db_ok = ('db' in context \
		and ('index' in context['db'] and dict == type (context['db']['index'])) \
		and ('languages' in context['db'] and dict == type (context['db']['languages'])) \
	)
	if not db_ok:
		holz.warn ('It appears db is corrupt.')
	repo_ok = ('repo' in context)
	if not repo_ok:
		holz.warn ('Remote repository info appears corrupt.')

	return paths_ok \
		and  db_ok \
		and repo_ok


def context_create (paths, repo_info):
	"""! Creates context map object.

	Loads and parses whistle JSON databases (raw index and language lookup).

	Returs a map containing:
	
	* paths: Data paths. See @ref "data_paths ()".
	* db: Voice and languages information. Contains following keys:
		* index: Voice index fetched from huggingface repository.
		* language: Generated language lookup to select voices by language code.
	* repo: Repo config. See @ref "remote_repo_config ()".
	
	@param paths Paths map. Can be obtained via @ref "data_paths ()".
	@param repo_info Remote repo information map. Can be obtained via @ref "remote_repo_config ()".

	@return Returns voice and language lookups.
	"""
	db = {
		'index': None,
		'languages': None
	}

	if not os.path.exists (paths['index']):
		holz.error ('No database index found!')
	else:
		with open (paths['index'], 'r') as f:
			db['index'] = json.load (f)

	if not os.path.exists (paths['languages']):
		holz.error ('No language lookup found!')
	else:
		with open (paths['languages'], 'r') as f:
			db['languages'] = json.load (f)

	lgl_path = os.path.join (paths['data'], 'legal.json')
	if not os.path.exists (lgl_path):
		holz.error ('No legal lookup found!')
	else:
		with open (lgl_path, 'r') as f:
			db['legal'] = json.load (f)

	context = {
		'paths': paths,
		'db': db,
		'repo': repo_info
	}

	if not _context_is_valid (context):
		return None
	
	return context


def context_guess_language_from_name (context, needle):
	"""! Searches the index and language lookup for a language which comes closest to your query.

	The search is done via fuzzy matching, trying to find the closest resemblance
	of the language or country name you provided.

	@param context Context map containig whistle index and language info. Can be created via @ref "context_create ()".
	@param needle Search term. This will be matched against countries, language names, codes.

	@return Returns the corresponding country code, or None if no match could be found.
	"""
	langdb = context['db']['languages']
	matching_code = None	

	for lang_code in langdb:
		matched, c = search.looks_like (needle, lang_code, 0.91, False)
		if matched:
			holz.info (f'{needle} ~ {lang_code}')
			matching_code = lang_code
			break
		lang_name_native = langdb[lang_code]['name_native']
		matched, c = search.looks_like (needle, lang_name_native, 0.8)
		if matched:
			holz.info (f'{needle} ~ {lang_name_native} (native)')
			matching_code = lang_code
			break
		lang_name_en = langdb[lang_code]['name_english']
		matched, c = search.looks_like (needle, lang_name_en, 0.8)
		if matched:
			holz.info (f'{needle} ~ {lang_name_en} (english)')
			matching_code = lang_code
			break
		lang_country = langdb[lang_code]['country_english']
		matched, c = search.looks_like (needle, lang_country, 0.7)
		if matched:
			matching_code = lang_code
			holz.info (f'{needle} ~ {lang_country}')
			break

	holz.info (matching_code)

	return matching_code


def model_list_installed (paths):
	"""! Searches piper-whistle cache for all models installed.

	Checks the user cache path (i.e. ~/.config/piper-whistle on *nix)

	The returned list contains map object with the following keys:
	
	' name: Clean voice name.
	' quality: Voice quality.
	' code: Language / Country code. (e.g. en_GB)
	' path: The absolute path to the onnx model.

	@param context Context map containig whistle index and language info. Can be created via @ref "context_create ()".

	@return Returns a list containing all installed models.
	"""
	models = []
	p = pathlib.Path (paths['voices'])
	for lang_dir in p.iterdir ():
		code = lang_dir.name
		for voice_dir in lang_dir.iterdir ():
			onnx_file_path = voice_dir.joinpath (f'{voice_dir.name}.onnx')
			if not onnx_file_path.exists ():
				continue

			# name has language code prepended. remove.
			rest = voice_dir.name.replace (f'{code}-', '')
			# now split off quality
			voice_name, voice_quality = rest.split ('-')
			models.append ({
				'name': voice_name,
				'quality': voice_quality,
				'code': code,
				'path': onnx_file_path.as_posix ()
			})

	return models


def model_resolve_path (paths, model_info):
	"""! Searches piper-whistle cache for a model corresponding to given specs.

	Checks the user cache path (i.e. ~/.config/piper-whistle on *nix) for a model
	constraied by given name, quality and speaker.

	The returned list contains map object with the following keys:
	
	* name: Clean voice name.
	* quality: Voice quality.
	* code: Language / Country code. (e.g. en_GB)
	* path: The absolute path to the onnx model.

	@param paths Paths map. Can be obtained via @ref "data_paths ()".
	@param model_info Map containing name, quality and speaker.

	@return Returns the path to the model, or None if nothing is found.
	"""
	name = model_info['name']
	quality = model_info['quality']
	speaker = model_info['speaker']

	voice_file_path = None
	p = pathlib.Path (paths['voices'])
	for lang_dir in p.iterdir ():
		code = lang_dir.name
		for voice_dir in lang_dir.iterdir ():
			# name has language code prepended. remove.
			rest = voice_dir.name.replace (f'{code}-', '')
			# now split off quality
			voice_name, voice_quality = rest.split ('-')
			if voice_name == name and voice_quality == quality:
				filename = f'{voice_dir.name}.onnx'
				voice_file_path = voice_dir.joinpath (filename).as_posix ()
				break

	return voice_file_path


def model_remove (paths, model_info):
	"""! Removes given model from piper-whistle cache.

	@return Returns the true if remove, false otherwise.
	"""
	model_path = model_resolve_path (paths, model_info)
	if not model_path:
		holz.warn (f'Could not find model with name "{model_info["name"]}@{model_info["quality"]}".')
		return False

	p = pathlib.Path (model_path)
	pr = p.resolve ().parent
	
	for model_part in pr.iterdir ():
		holz.debug (f'Removing "{model_part}" ...')
		model_part.unlink ()

	holz.debug (f'Removing "{pr}" ...')
	pr.rmdir ()

	sys.stdout.write (f'Removed "{model_info["name"]}@{model_info["quality"]}".\n')

	return True



"""CLI functions corresponding to available commands.

Defines functions to handle available CLI commands.
```
guess: run_guess
path: run_path
speak: run_speak
list: run_list
preview: run_preview
install: run_install
remove: run_remove
```
Command handle function get passed whistle context and argparse arguments.

"""
# 2023-∞ (c) blurryroots innovation qanat OÜ. All rights reserved.
import subprocess
import os
import sys
import json
import requests
import userpaths
import time
import urllib
import pathlib
# Append root package to path so it can be called with absolute path.
sys.path.append (str (pathlib.Path(__file__).resolve().parents[1]))
from piper_whistle import holz
from piper_whistle import db
from piper_whistle import util


def _run_program (params):
	"""! Runs external program with given parameters
	@param params List of parameters to use for program invokation.
	@return Returns a touple containing (returncode, stdout, stderr)
	"""
	holz.debug (f'Running external program: {params}')
	p = subprocess.Popen (params
		, bufsize=0
		, stdout=subprocess.PIPE
		,  stderr=subprocess.PIPE
	)

	out, err = p.communicate ()
	holz.debug (f'Exit with {p.returncode}')
	return (p.returncode, out, err)


def _parse_voice_selector (selector):
	"""! Get voice details from selector string.
	@param selector Voice identifying string.
	@return Returns a touple containgin (name, quality, speaker) of the voice.
	"""
	if '-' in selector:
		code, name, quality = selector.split ('-')
		return name, quality, 0

	name, rest = selector.split ('@')
	if '/' in rest:
		quality, speaker = rest.split ('/')
	else:
		quality = rest
		speaker = 0

	return name, quality, speaker


def run_refresh (context, args):
	"""! Run command 'refresh'
	@param context Context information and whistle database.
	@param args Processed arguments (prepared by argparse).
	@return Returns 0 on success, otherwise > 0.
	"""
	if 'refresh' == args.command:
		context['repo']['repo-id'] = args.repository
	else:
		holz.warn (
			f'-R will be phased out within the next couple releases. '
			f'Please use the "refresh" command.'
		)

	holz.info (f'Fetching and rebuilding database from "{context["repo"]["repo-id"]}" ...')
	context = db.index_download_and_rebuild (context['paths'], context['repo'])
	if not context:
		holz.info (f'Could not rebuild index.')
		return 13

	holz.debug (f"Noting last update time to '{context['paths']['last-updated']}'.")
	with open (context['paths']['last-updated'], 'r') as f:
		sys.stdout.write (f.read ())

	return 0


def run_guess (context, args):
	"""! Run command 'guess'
	@param context Context information and whistle database.
	@param args Processed arguments (prepared by argparse).
	@return Returns 0 on success, otherwise > 0.
	"""
	target_lang = args.language_name
	matching_code = db.context_guess_language_from_name (context, target_lang)

	if matching_code:
		lang = context['db']['languages'][matching_code]
		holz.info (
			f'Best guess for "{target_lang}": '
			f'{matching_code} ({lang["name_native"]} [{lang["name_english"]}])'
		)
		sys.stdout.write (matching_code)
		return 0

	holz.error (f'Could not find anything matching: {target_lang}')
	return 13


def run_path (context, args):
	"""! Run command 'path'
	@param context Context information and whistle database.
	@param args Processed arguments (prepared by argparse).
	@return Returns 0 on success, otherwise > 0.
	"""
	selector = args.voice_selector

	code = None
	if ':' in selector:
		code, selector = selector.split (':')

	name, quality, speaker = _parse_voice_selector (selector)

	model_info = {
		'name': name,
		'quality': quality,
		'speaker': speaker
	}
	voice_file_path = db.model_resolve_path (context['paths'], model_info)
	if not voice_file_path:
		holz.error (f'Could not find any voice matching {name}!')
		return 13

	sys.stdout.write (voice_file_path)
	return 0


def run_speak (context, args):
	"""! Run command 'speak'
	@param context Context information and whistle database.
	@param args Processed arguments (prepared by argparse).
	@return Returns 0 on success, otherwise > 0.
	"""
	p = pathlib.Path (args.channel)

	if not p.exists ():
		holz.error (f'No channel at "{p}" found.')
		return 13

	with open (p.as_posix (), 'w') as f:
		payload = f'{args.something}'
		if args.json:
			j = {'text': payload}
			if args.output:
				# If relative path, prepend current working directory.
				op = pathlib.Path (args.output)
				if not op.is_absolute ():
					op = pathlib.Path.cwd ().joinpath (args.output)
				j['output_file'] = op.absolute ().as_posix ()
			payload = json.dumps (j)
		
		payload = payload + '\n'

		holz.info (f'Sending {len (payload)} bytes ...')
		r = f.write (payload)
		holz.info (f'{r} bytes sent.')

	return 0


def run_list (context, args):
	"""! Run command 'list'
	@param context Context information and whistle database.
	@param args Processed arguments (prepared by argparse).
	@return Returns 0 on success, otherwise > 0.
	"""
	langdb = context['db']['languages']

	if args.languages:
		available_lang_codes = '\n'.join ([c for c in langdb])
		sys.stdout.write (f'{available_lang_codes}')
		return 0

	if args.installed:
		voice_i = 0
		for model in db.model_list_installed (context['paths']):
			code = model['code']
			sys.stdout.write (
				f"\t{code}:{model['name']}@{model['quality']}"
			)
			if args.verbose:
				sys.stdout.write (f"\t{model['path']}")
			if args.legal:
				lgl = context['db']['legal'][model['name']]
				a = f"Voice[{lgl['training']}]: {lgl['license']}, " \
					f"Reference: {lgl['reference']}, " \
					f"Dataset: {lgl['dataset-url']}"
				sys.stdout.write (f"\t{a}")
			if args.show_url:
				download_info = db.assemble_download_info (context
					, code
					, voice_i
				)
				sys.stdout.write (f"\t{download_info['model']['url']}")
			sys.stdout.write ("\n")
			voice_i += 1

		return 0
	
	if args.all:
		for code in langdb:
			lang = langdb[code]

			sys.stdout.write (f'Voices for "{code}":\n')
			voice_i = 0
			for voice_name in lang['voices']:
				details = voice_name
				if args.legal:
					lgl = context['db']['legal'][voice_name]
					a = f"Voice[{lgl['training']}]: {lgl['license']}, " \
						f"Reference: {lgl['reference']}, " \
						f"Dataset: {lgl['dataset-url']}"
					details = f"{details}\t({a})"
				if args.show_url:
					download_info = db.assemble_download_info (context
						, code
						, voice_i
					)
					details = f"{a}\t{download_info['model']['url']}"
				sys.stdout.write (f"\t{voice_i}: {details}\n")

				voice_i = voice_i + 1

		return 0

	code = args.language_code
	if not code in langdb:
		holz.error (f'Could not find language with code "{code}".')
		return 13

	lang = langdb[code]
	voice_i = args.voice_index
	index = context['db']['index']

	if -1 < voice_i:
		voice_name = lang['voices'][voice_i]

		sys.stdout.write (f'{voice_name}\t{voice_i}')
		if args.legal:
			lgl = context['db']['legal'][voice_name]
			a = f"Voice[{lgl['training']}]: {lgl['license']}, "\
				f"Reference: {lgl['reference']}, " \
				f"Dataset: {lgl['dataset-url']}"
			sys.stdout.write (f'\t{a}')
		if args.show_url:
			download_info = db.assemble_download_info (context
				, code
				, voice_i
			)
			sys.stdout.write (f"\t{download_info['model']['url']}")
		
		sys.stdout.write ('\n')

		if not args.omit_speakers:
			speakers = index[voice_name]['speaker_id_map']

			sys.stdout.write (f'Speakers:\n')
			if 0 == len (speakers):
				sys.stdout.write (f'\t\t0 (no-name)')
			else:
				for speaker_id in speakers:
					sys.stdout.write (f'\t\t{speakers[speaker_id]:3} ({speaker_id})')
	else:
		sys.stdout.write (f'Available voices ({code}):\n')
		voice_i = 0
		for voice_name in lang['voices']:
			sys.stdout.write (f"\t{voice_i}: {voice_name}")
			if args.legal:
				lgl = context['db']['legal'][voice_name]
				a = f"Voice[{lgl['training']}]: {lgl['license']}, " \
					f"Reference: {lgl['reference']}, " \
					f"Dataset: {lgl['dataset-url']}"
				sys.stdout.write (f'\t{a}')
			if args.show_url:
				download_info = db.assemble_download_info (context
					, code
					, voice_i
				)
				sys.stdout.write (f"\t{download_info['model']['url']}")
			sys.stdout.write ('\n')

			voice_i += 1

	return 0


def run_preview (context, args):
	"""! Run command 'preview'
	@param context Context information and whistle database.
	@param args Processed arguments (prepared by argparse).
	@return Returns 0 on success, otherwise > 0.
	"""
	dry_run = args.dry_run
	code = args.language_code
	voice_i = args.voice_index
	speaker_i = args.speaker_index
	index = context['db']['index']
	langdb = context['db']['languages']
	
	holz.info (f'Looking up preview info for {code}:{voice_i}:{speaker_i} ...')
	speaker_url = None
	download_info = db.assemble_download_info (context, code, voice_i)
	if download_info:
		if -1 < speaker_i and speaker_i < len (download_info['samples']):
			speaker_url = download_info['samples'][speaker_i]

	holz.debug (download_info)

	if not speaker_url:
		holz.error ('Could not find sample for this voice.')
		return 13

	p = pathlib.Path (f"{context['paths']['voices']}")
	p = p.joinpath (f"{download_info['local_path_relative']}")
	p = p.joinpath (f"{local_voice_path}/samples")
	p = p.resolve ()

	p.mkdir (parents=True, exist_ok=True)
	
	filename = os.path.basename (speaker_url)
	file_path = p.joinpath (filename)
	if file_path.exists ():
		holz.info ('Cached file detected.')
	else:
		holz.info (f'Downloading ({speaker_url}) ...')
		if dry_run:
			sys.stdout.write (f'dl {speaker_url} > {file_path.as_posix ()}\n')
		else:
			r = util.download_as_stream_with_progress (
				speaker_url, file_path.as_posix ()
			)
			if 0 > r:
				holz.error ('Error downloading model.')
				return 13

	sys.stderr.write (f'Playing {file_path.as_posix ()} ...\n')
	if dry_run:
		sys.stdout.write (f'Dry-run, skipping play.')
	else:
		if not file_path.exists ():
			holz.error ('File not found.')
			return 13

		# Suppress playsound's warning message about pygobject.
		# Has to be done before loading modulde, since the loading code
		# is producing the log message.
		import logging
		holz.setup ('playsound', logging.ERROR)

		# After setup, load the module and play the sound.
		from playsound import playsound
		# TODO: On linux, this spins up another python instance, which is 
		# not the most convenient way in terms of stopping / cancelling.
		playsound (file_path.as_posix (), block = True)

		holz.info (f'Finished with: 0')

	return 0


def run_install (context, args):
	"""! Run command 'install'
	@param context Context information and whistle database.
	@param args Processed arguments (prepared by argparse).
	@return Returns 0 on success, otherwise > 0.
	"""
	dry_run = args.dry_run
	code = args.language_code
	voice_i = args.voice_index
	index = context['db']['index']
	langdb = context['db']['languages']

	download_info = db.assemble_download_info (context, code, voice_i)
	if not download_info:
		holz.error ('Could not find any downloads for this configuration.')
		return 13
	
	p = pathlib.Path (f"{context['paths']['voices']}")
	p = p.joinpath (f"{download_info['local_path_relative']}")
	p = p.resolve ()

	p.mkdir (parents=True, exist_ok=True)
	holz.info (f'Using voice path at: {p}')

	config_url = download_info['config']['url']
	config_file_path = p.joinpath (os.path.basename (config_url))
	if not config_file_path.exists ():
		size = util.float_round (float (download_info['config']['size']) / 1024)
		holz.info (f'Fetching config ({size}kb) ...')
		if not dry_run:
			r = util.download_as_stream_with_progress (config_url
				, config_file_path.as_posix ()
			)
			if 0 > r:
				holz.error ('Error downloading config.')
				return 13

	else:
		holz.info ('Config already cached.')

	model_url = download_info['model']['url']
	model_file_name = os.path.basename (model_url)
	model_name = model_file_name.split ('.')[0]
	model_file_path = p.joinpath (model_file_name)
	if not model_file_path.exists ():
		size = util.float_round (
			float (download_info['model']['size'])
			/ 1024
			/ 1024
		)
		holz.info (f'Fetching {model_name} ({size}mb) ...')
		if not dry_run:
			r = util.download_as_stream_with_progress (
				model_url, model_file_path.as_posix ()
			)
			if 0 > r:
				holz.error ('Error downloading model.')
				return 13
	else:
		holz.info ('Model already cached.')

	selector = download_info["selection_name"]
	sys.stdout.write (f'{selector}\t{model_name}\t{model_file_path}')

	return 0


def run_remove (context, args):
	"""! Run command 'remove'
	@param context Context information and whistle database.
	@param args Processed arguments (prepared by argparse).
	@return Returns 0 on success, otherwise > 0.
	"""
	selector = args.voice_selector

	code = None
	if ':' in selector:
		code, selector = selector.split (':')

	name, quality, speaker = _parse_voice_selector (selector)

	model_info = {
		'name': name,
		'quality': quality,
		'speaker': speaker
	}
	did_remove = db.model_remove (context['paths'], model_info)
	if not did_remove:
		holz.error (f'Could not remove "{name}@{quality}"!')
		return 13

	return 0



"""
Handle version related information.
"""

# Define in common three component form.
MAJOR = 1
MINOR = 6
PATCH = 183


def as_list ():
	'''Returns the three components as list'''
	return [MAJOR, MINOR, PATCH]


def as_string ():
	'''Joins the version components separated by a . (dot)'''
	return ".".join ([str (comp) for comp in as_list ()])


if '__main__' == __name__:
	print (as_string ())



"""Utility functions.
"""
# 2023-∞ (c) blurryroots innovation qanat OÜ. All rights reserved.
import sys
import pathlib
import requests
from tqdm import tqdm
# Append root package to path so it can be called with absolute path.
sys.path.append (str (pathlib.Path(__file__).resolve().parents[1]))
# ..
from piper_whistle import holz


def float_round (x, digits = 2):
	"""! Rounds given number to a specific precision
	@param x Number to round.
	@param digits The number of significant digitis to retain. Default is 2.
	@return Returns x rounded to significant digits.
	"""
	k = 10**digits
	xx = int (x * k)
	x = float (xx) / k
	return x


def download_as_stream_with_progress (url: str, file_path: str, label: str = 'dl'):
	"""! Downloads a resource via streaming get request and shows progress.
	@param url Target URL to download.
	@param file_path Path to store downloaded file to.
	@param label Annotation used when printing progress. 'dl' by default.
	@return Returns number of bytes downloaded, or -1 on error.
	"""
	resp = requests.get (url, stream = True)
	if not (300 > resp.status_code):
		holz.error (f'Could not fetch "{url}"! (c: {resp.status_code})')
		try:
			import rich
			rich.inspect (r)
		except:
			holz.error (
				f'Skipping verbose response logging. '
				f'Module "rich" not installed.'
			)

		return -1

	total = int (resp.headers.get ('content-length', 0))
	# Can also replace 'file' with a io.BytesIO object
	with open (file_path, 'wb') as file, tqdm (
		desc = label,
		total = total,
		unit = 'iB',
		unit_scale = True,
		unit_divisor = 1024,
	) as bar:
		for data in resp.iter_content (chunk_size = 1024):
			size = file.write (data)
			bar.update (size)

	return total



#!env python3
"""    
	setuptools script to install / build piper-whistle.

	2021-∞ (c) blurryroots innovation qanat OÜ. All rights reserved.
	See license.md for details.

	https://think-biq.com
"""

import setuptools
from src.piper_whistle import version

with open("readme.md", "r") as fh:
	long_description = fh.read()

setuptools.setup (
	name = "piper_whistle",
	version = version.as_string (),
	author = "biq",
	author_email = "sven.freiberg@biq.solutions",
	description = "CLI tool to manage piper voices.",
	long_description = long_description,
	long_description_content_type = "text/markdown",
	keywords="text-to-speech piper voice model tool",
	url = "https://gitlab.com/think-biq/piper-whistle",
	package_dir = { 'piper_whistle': 'src/piper_whistle' },
	packages = ['piper_whistle'],
	classifiers = [
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	python_requires = '>=3.8',
	entry_points = {
		'console_scripts': [
			'piper_whistle = piper_whistle.cli:main_sys'
		],
	},
	install_requires = [
		'attrs==23.1.0',
		'certifi==2023.7.22',
		'charset-normalizer==3.3.2',
		'filelock==3.13.1',
		'fsspec==2023.10.0',
		'huggingface-hub==0.18.0',
		'idna==3.4',
		'mdurl==0.1.2',
		'playsound==1.3.0',
		'PyYAML==6.0.1',
		'rapidfuzz==3.5.1',
		'requests==2.31.0',
		'thefuzz==0.20.0',
		'tqdm==4.66.1',
		'typing_extensions==4.8.0',
		'urllib3==2.0.7',
		'userpaths==0.1.3',
	]
)